from wagtail import blocks
from django.utils.safestring import mark_safe
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from .forms import (
    AceEditorField,
    AceEditorValue,
)

class AceEditorBlock(blocks.FieldBlock):
    def __init__(
        self,
        mode="ace/mode/django",
        theme="ace/theme/wagtail",
        include_template_context: bool  = False,
        use_frame_preview: bool         = False,
        frame_css: list[str]            = None,
        frame_js: list[str]             = None,
        clean_html: bool                = False,
        disable_preview: bool           = False,
        preview_checkbox_checked: bool  = False,
        required: bool                  = True,
        help_text: str                  = None,
        validators=(),
        **kwargs,
    ):
        self.field_options = {
            "required": required,
            "help_text": help_text,
            "validators": validators,
            "mode": mode,
            "theme": theme,
            "clean_html": clean_html,
            "use_frame_preview": use_frame_preview,
            "preview_checkbox_checked": preview_checkbox_checked,
            "disable_preview": disable_preview,
            "frame_css": frame_css or [],
            "frame_js": frame_js or [],
        }

        self.include_template_context = include_template_context

        super().__init__(**kwargs)

    class Meta:
        icon = "doc-full"
        label = _("Ace Editor")

    @cached_property
    def field(self):
        return AceEditorField(**self.field_options)
    
    def to_python(self, value):

        if value is None:
            return None
        
        if isinstance(value, AceEditorValue):
            return value

        return AceEditorValue(
            value,
            mode=self.field_options["mode"],
            theme=self.field_options["theme"],
            clean_html=self.field_options["clean_html"],
            render_with_context=self.include_template_context,
        )
    




from typing import Any
from django import forms
from django.utils.safestring import mark_safe
from django.template import Template, Context

from .widgets import AceEditorWidget


class AceEditorValue(str):
    def __new__(cls, value, mode="ace/mode/django", theme="ace/theme/wagtail", clean_html=False, render_with_context=False):
        self = super().__new__(cls, value)
        self.mode = mode
        self.theme = theme
        self.clean_html = clean_html
        self.render_with_context = render_with_context
        return self
    
    def get_context(self, context = None):
        if context is None:
            context = {}
        
        return {
            "self": self,
            "mode": self.mode,
            "theme": self.theme,
            **context
        }
    
    def render_as_block(self, context=None):
        tpl = self
        if self.render_with_context:
            if context is None:
                new_context = self.get_context(context=context)
            else:
                new_context = self.get_context(context=dict(context))

            tpl = Template(tpl)
            return tpl.render(Context(new_context))

        return mark_safe(tpl)


class AceEditorField(forms.CharField):

    def __init__(self, mode="ace/mode/django", theme="ace/theme/wagtail", use_frame_preview=False, frame_css=None, frame_js=None, clean_html=False, disable_preview=False, include_template_context=False, preview_checkbox_checked=True, *args, **kwargs):
        self.mode = mode
        self.theme = theme
        self.use_frame_preview = use_frame_preview
        self.frame_css = frame_css
        self.frame_js = frame_js
        self.clean_html = clean_html
        self.disable_preview = disable_preview
        self.include_template_context = include_template_context
        self.preview_checkbox_checked = preview_checkbox_checked
        self._widget = None
        super().__init__(*args, **kwargs)

    @property
    def widget(self):
        if self._widget:
            return self._widget
        
        return AceEditorWidget(
            mode=self.mode,
            theme=self.theme,
            use_frame_preview=self.use_frame_preview,
            frame_css=self.frame_css,
            frame_js=self.frame_js,
            clean_html=self.clean_html,
            disable_preview=self.disable_preview,
            preview_checkbox_checked=self.preview_checkbox_checked
        )
    
    def to_python(self, value: Any | None) -> Any | None:

        if value is None:
            return None
        
        if isinstance(value, AceEditorValue):
            return value

        return AceEditorValue(
            value,
            mode=self.mode,
            theme=self.theme,
            clean_html=self.clean_html,
            render_with_context=self.include_template_context
        )
    
    @widget.setter
    def widget(self, value):
        self._widget = value



    


from typing import Any
from django.forms import widgets
from django.templatetags.static import static
from .clean import (
    BLEACH_CLEAN_ACE_MODES,
    clean_html,
)
import json, re



_proto_re = re.compile(r'^(?:[a-z+]+:)?//', re.I)


def format_static_if_needed(value):
    if not value:
        return value
    
    if isinstance(value, str):
        
        if _proto_re.match(value):
            return value
        
        return static(value)

    return [format_static_if_needed(v) for v in value]


class AceEditorWidget(widgets.Textarea):
    template_name = "wagtail_ace_editor/ace_editor.html"

    def __init__(self, mode="ace/mode/django", theme="ace/theme/wagtail", use_frame_preview=True, frame_css=None, frame_js=None, clean_html=False, disable_preview=False, preview_checkbox_checked=True, attrs=None):
        attrs                         = attrs or {}
        self.mode                     = mode
        self.theme                    = theme
        self.disable_preview          = True
        self.preview_checkbox_checked = preview_checkbox_checked
        self.use_frame_preview        = use_frame_preview and not disable_preview
        self.frame_css                = frame_css or [
            "wagtail_ace_editor/css/ace-editor-iframe.css"
        ] if use_frame_preview else []
        self.frame_js                 = frame_js or []
        self.clean_html               = clean_html
        super().__init__(attrs=attrs)

    def format_value(self, value) -> str | None:
        value = super().format_value(value)
        if value is None:
            return ""
        return value
    
    def value_from_datadict(self, data, files, name):
        value = super().value_from_datadict(data, files, name)
        if self.clean_html and self.mode.lower() in BLEACH_CLEAN_ACE_MODES:
            return clean_html(value)
        return value

    def build_attrs(self, *args, **kwargs):
        attrs = super().build_attrs(*args, **kwargs)
        mode = self.mode
        if callable(mode):
            mode = mode()

        theme = self.theme
        if callable(theme):
            theme = theme()
            
        attrs["data-controller"] = "ace-editor"
        attrs["data-ace-editor-mode-value"] = mode
        attrs["data-ace-editor-theme-value"] = theme
        attrs["data-ace-editor-use-frame-preview-value"] = self.use_frame_preview
        attrs["data-ace-editor-frame-css-value"] = json.dumps(self.build_frame_css())
        attrs["data-ace-editor-frame-js-value"] = json.dumps(self.build_frame_js())
        attrs["data-ace-editor-preview-checkbox-checked-value"] = self.preview_checkbox_checked
        return attrs
    
    def get_context(self, name: str, value: Any, attrs: dict[str, Any] | None) -> dict[str, Any]:
        return super().get_context(name, value, attrs) | {
            "disable_preview": self.disable_preview,
        }
    
    def build_frame_css(self):
        return format_static_if_needed(self.frame_css)

    def build_frame_js(self):
        return format_static_if_needed(self.frame_js)

    class Media:
        js = [
            "wagtail_ace_editor/js/ace.js",
            "wagtail_ace_editor/js/ace-editor.js",
            "wagtail_ace_editor/js/ace-editor-controller.js"
        ]




from django.test import TestCase, Client
from django.conf import settings
from django import forms
from os import path
import re
from wagtail_ace_editor.forms import AceEditorField
from wagtail.templatetags.wagtailcore_tags import (
    include_block,
)


FIXTURES = path.join(settings.BASE_DIR, "core/fixtures")

space_re = re.compile(r"^\s+", re.MULTILINE)


class WidgetForm(forms.Form):
    test_field = AceEditorField(
        mode="ace/mode/django",
        theme="ace/theme/wagtail",
        use_frame_preview=False,
        frame_css=[],
        frame_js=[],
        include_template_context=False,
        required=False,
    )

class WithContextWidgetForm(forms.Form):
    test_field_tpl_context = AceEditorField(
        mode="ace/mode/django",
        theme="ace/theme/wagtail",
        use_frame_preview=False,
        include_template_context=True,
        required=False,
    )

class WidgetTestCase(TestCase):

    def test_widget_rendering(self):
        form = WidgetForm()
        p = form.as_p()
        p = space_re.sub(' ', p)
        print(p)
        self.assertInHTML("""<p>
<label for="id_test_field">Test field:</label>
<div class="ace-editor-widget">
<div class="ace-editor-preview-wrapper" id="id_test_field-preview-wrapper"></div>
<div id="id_test_field-wrapper" class="ace-editor-widget-wrapper">
<textarea id="id_test_field" name="test_field" ace-editor-input cols="40"rows="10"id="id_test_field"data-controller="ace-editor"data-ace-editor-mode-value="ace/mode/django"data-ace-editor-theme-value="ace/theme/wagtail"data-ace-editor-use-frame-preview-value="False"data-ace-editor-frame-css-value="[]"data-ace-editor-frame-js-value="[]"></textarea>
<div id="id_test_field-editor"></div>
<div class="ace-editor-widget-preview-button">
<label for="id_test_field-preview-text">Preview:</label>
<input type="checkbox" id="id_test_field-preview-check">
</div>
</div>
</div>
</p>""", p)
        
    def test_to_python(self):
        original = "<p>test {{ with_context }}</p>"
        form = WidgetForm({
            "test_field": original,
        })
        self.assertTrue(form.is_valid())
        test_field = form.cleaned_data["test_field"]
        rendered = test_field.render_as_block({
            "without_context": "context value",
        })
        self.assertEqual(rendered, original)
        self.assertEqual(str(test_field), original)

        
    def test_to_python_tpl_context(self):
        original = "<p>test {{ with_context }}</p>"
        form = WithContextWidgetForm({
            "test_field_tpl_context": original,
        })
        self.assertTrue(form.is_valid())
        test_field = form.cleaned_data["test_field_tpl_context"]
        rendered = test_field.render_as_block({
            "with_context": "context value",
        })
        self.assertEqual(rendered, "<p>test context value</p>")
        self.assertEqual(str(test_field), original)

